import 'package:after_layout/after_layout.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';

import '../../osam_flutter.dart';

PageRoute _pageRoute({WidgetBuilder widgetBuilder, RouteSettings settings, bool isMaterial = false}) => isMaterial
    ? MaterialPageRoute(builder: widgetBuilder, settings: settings)
    : CupertinoPageRoute(builder: widgetBuilder, settings: settings);

class OsamNavigator<P extends NavigationPresenter> extends StatefulWidget {
  final Map<String, WidgetBuilder> _screens;
  final bool rootNavigator;
  final bool isMaterial;

  const OsamNavigator(this._screens, {Key key, this.rootNavigator = false, this.isMaterial = false}) : super(key: key);

  @override
  _OsamNavigatorState createState() => _OsamNavigatorState<P>();
}

class _OsamNavigatorState<P extends NavigationPresenter> extends State<OsamNavigator> with AfterLayoutMixin {
  final List<String> _routes = [];

  Map<String, WidgetBuilder> get screens => widget._screens;

  @override
  Widget build(BuildContext context) => screens[context.presenter<P>().routesStream.value.first](context);

  void _silentPop(NavigationPresenter presenter) {
    if (presenter.routesStream.value.length > 1) {
      _routes.removeLast();
      presenter.pop();
    }
  }

  @override
  void afterFirstLayout(BuildContext context) {
    final navigator = Navigator.of(context, rootNavigator: widget.rootNavigator);
    final presenter = context.presenter<P>();
    final initialRoutes = presenter.routesStream.value;
    _routes.addAll(initialRoutes);
    if (initialRoutes.length > 1) {
      final others = initialRoutes.sublist(1, initialRoutes.length);
      others.forEach((route) {
        navigator.push(_pageRoute(widgetBuilder: screens[route], isMaterial: widget.isMaterial)
          ..popped.then((_) => _silentPop(presenter)));
      });
    }
    presenter.routesSub.onData((newRoutes) {
      final routesDifferentNumber = newRoutes.length - _routes.length;
      if (routesDifferentNumber == 0 && newRoutes.last != _routes.last) {
        navigator.pushReplacement(_pageRoute(widgetBuilder: screens[newRoutes.last], isMaterial: widget.isMaterial)
          ..popped.then((_) => _silentPop(presenter)));
      } else if (routesDifferentNumber > 0) {
        final pushRoutes = newRoutes.sublist(_routes.length, newRoutes.length);
        pushRoutes.forEach((route) {
          navigator.push(_pageRoute(widgetBuilder: screens[route], isMaterial: widget.isMaterial)
            ..popped.then((_) => _silentPop(presenter)));
        });
      } else if (routesDifferentNumber < 0) {
        _routes.first != newRoutes.first
            ? navigator.pushAndRemoveUntil(
                _pageRoute(widgetBuilder: screens[newRoutes.first], isMaterial: widget.isMaterial), (route) => false)
            : _routes.sublist(newRoutes.length, _routes.length).forEach((route) {
                navigator.pop();
              });
      }
      _routes.clear();
      _routes.addAll(newRoutes);
    });
  }
}

class OsamTabNavigator<P extends TabNavigationPresenter> extends StatelessWidget {
  final Map<String, PresenterProvider<NavigationPresenter>> navigators;
  final Map<String, Widget> tabs;
  final BoxDecoration tabDecoration;
  final _navigatorKeys = <String, GlobalKey<NavigatorState>>{};

  OsamTabNavigator({Key key, this.navigators, this.tabDecoration, this.tabs}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    final presenter = context.presenter<P>();
    if (_navigatorKeys.isEmpty) {
      _navigatorKeys.addAll(navigators.map((k, v) => MapEntry(k, GlobalKey<NavigatorState>())));
    }
    return OsamBuilder<String>(
      key: const ValueKey('tab_ob'),
      stream: presenter.tabStream,
      builder: (ctx, snapshot) {
        return Stack(
          children: <Widget>[
            IndexedStack(
              index: navigators.keys.toList().indexOf(snapshot.data),
              children: navigators.keys
                  .map((tab) => _OsamNavigatorPage(
                        _navigatorKeys[tab],
                        navigators[tab],
                        key: ValueKey(tab.toString() + '_onp'),
                      ))
                  .toList(),
            ),
            Positioned(
              bottom: 0,
              child: SafeArea(
                bottom: true,
                child: Container(
                  decoration: tabDecoration,
                  padding: const EdgeInsets.symmetric(vertical: 5),
                  constraints: BoxConstraints(
                    maxWidth: MediaQuery.of(context).size.width,
                    minHeight: 48,
                  ),
                  child: Row(
                    children: navigators.keys
                        .map((tab) => Expanded(
                              child: GestureDetector(
                                child: Container(
                                  child: tabs[tab],
                                  alignment: Alignment.center,
                                  color: Colors.transparent,
                                ),
                                onTap: () {
                                  if (snapshot.data != tab) {
                                    presenter.changeTab(tab);
                                  } else {
                                    final currentPresenter = navigators[snapshot.data].presenter;
                                    final routes = currentPresenter.routesStream.value;
                                    if (routes.length > 1) {
                                      currentPresenter.pushAndRemoveUntil(routes.first);
                                    }
                                  }
                                },
                              ),
                            ))
                        .toList(),
                  ),
                ),
              ),
            )
          ],
        );
      },
    );
  }
}

class _OsamNavigatorPage extends StatelessWidget {
  final GlobalKey<NavigatorState> navKey;
  final Widget child;

  const _OsamNavigatorPage(this.navKey, this.child, {Key key}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return Navigator(
      key: navKey,
      initialRoute: '~/',
      onGenerateRoute: (RouteSettings settings) =>
          _pageRoute(settings: const RouteSettings(name: '~/'), widgetBuilder: (BuildContext context) => child),
    );
  }
}
